import secrets
from smtplib import SMTPException
from django.db import models
from django.utils import timezone
from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, BaseUserManager
from django.contrib.auth.hashers import make_password
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError
from django.core.signing import Signer
from django.core.mail import send_mail
from django.template.loader import get_template
from django.urls import reverse

SALT = 'W3VIvFmvhhvePbCmC_hcixstdpNRrjB5QqleEwvE2Uh00Mbi9VYfU-dixweZ3n1-e9JpkcHIw6XzCIZ_nefLqg'

# Helper functions for User creation notifications

def signed_url(new_user):
    return reverse(
        'vkk:users:registration_confirm',
        args=new_user.get_signed_token().split('/')
    )

def send_registration_mail(user):
    """Notifies the User of their registration."""
    """A new NewUser instance is derived accordingly."""
    new_user = NewUser.objects.create(user=user)
    context = {'user': user, 'new_user': new_user, 'signed_url': signed_url(new_user)}
    subject = get_template('vkk/users/registration_subject.txt')
    message = get_template('vkk/users/registration_email.txt')
    try:
        send_mail(
            subject=subject.render(context=context),
            message=message.render(context=context),
            from_email=None,
            recipient_list=[user.email]
        )
    except SMTPException:
        new_user.delete()
        raise

def send_renew_mail(new_user):
    """Renews a NewUser instance and sends a notification to them."""
    user = new_user.user
    new_user.renew_token()
    context = {'user': user, 'new_user': new_user, 'signed_url': signed_url(new_user)}
    subject = get_template('vkk/users/renew_subject.txt')
    message = get_template('vkk/users/renew_email.txt')
    send_mail(
        subject=subject.render(context=context),
        message=message.render(context=context),
        from_email=None,
        recipient_list=[user.email]
    )

# Create your models here.

class UserManager(BaseUserManager):
    def _create_user(self, email, first_name, password, **extra_fields):
        """
        Creates and saves a user with the given email, password and first_name. (last_name as well, if provided.)
        """
        email = self.normalize_email(email)
        if not email or email == '':
            raise ValueError('An email address must be provided')
        user = self.model(email=email, first_name=first_name, **extra_fields)
        user.password = make_password(password)
        user.full_clean()
        user.save(using=self._db)
        return user

    def create_user(self, email, first_name, password=None, **extra_fields):
        """
        Creates and saves a typical user.
        """
        extra_fields.setdefault('is_accountant', False)
        extra_fields.setdefault('is_active', False)
        return self._create_user(email, first_name, password, **extra_fields)
    
    def create_superuser(self, email, first_name, password=None, **extra_fields):
        """
        Creates and saves a user with accounting permissions.
        """
        extra_fields.setdefault('is_accountant', True)
        extra_fields.setdefault('is_active', False)
        if extra_fields.get('is_accountant') is not True:
            raise ValueError('Accountants must have is_accounting=True')
        return self._create_user(email, first_name, password, **extra_fields)

class User(AbstractBaseUser):
    """
    An option user model similar to AbstractUser stripped to the needs of this app.
    """
    first_name = models.CharField(_('first name'), max_length=150)
    last_name = models.CharField(_('last name'), max_length=150, blank=True)
    email = models.EmailField(_('email address'), unique=True)
    is_accountant = models.BooleanField(
        verbose_name=_('accountant status'),
        default=False,
        help_text=_('Designates whether the user can access features reserved for accounting.'),
    )
    is_active = models.BooleanField(
        verbose_name=_('active'),
        default=True,
        help_text=_(
            'Designates whether this user should be treated as active. '
            'Unselect this instead of deleting accounts.'
        ),
    )

    EMAIL_FIELD = 'email'
    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = ['first_name']

    objects = UserManager()

    class Meta:
        verbose_name =_('user')
        verbose_name_plural = _('users')
        default_permissions = ()

    def get_full_name(self):
        """
        Return the first_name plus the last_name, with a space in between.
        """
        full_name = '%s %s' % (self.first_name, self.last_name)
        return full_name.strip()

    def get_reverse_full_name(self):
        """
        Return the last_name plus the first_name, with a space in between.
        """
        full_name = '%s, %s' % (self.last_name, self.first_name)
        return full_name.strip()

    def get_short_name(self):
        """Return the short name for the user."""
        return self.last_name

    def clean(self):
        super().clean()
        self.email = self.__class__.objects.normalize_email(self.email)

class NewUserManager(models.Manager):
    def create(self, **kwargs):
        token = secrets.token_urlsafe(32)
        while(self.get_queryset().filter(token=token).exists()):
            token = secrets.token_urlsafe(32)
        kwargs.setdefault('token', token)
        kwargs.setdefault(
            'expiration_date',
            timezone.now().date() + timezone.timedelta(days=21)
        )
        return super().get_queryset().create(**kwargs)

    # raises django.core.signing.BadSignature
    def get_with_signed_token(self, signed_token):
        """Return New User with signature check for token"""
        signer = Signer(sep='/', salt=SALT)
        token = signer.unsign(signed_token)
        return self.get_queryset().get(token=token)    

class NewUser(models.Model):
    """Model to keep track of user who have not yet confirmed their credentials."""
    user = models.OneToOneField(
        to=settings.AUTH_USER_MODEL,
        verbose_name=_('user'),
        on_delete=models.CASCADE
    )
    token = models.SlugField(
        verbose_name=_('token'),
        unique=True
    )
    expiration_date = models.DateField(
        verbose_name=_('expiration date'),
        help_text="Date after which the generated token will be expired."
    )

    def renew_token(self):
        """Renews the token and sets the expiration date accordingly."""
        self.token  = secrets.token_urlsafe(32)

        # making sure no duplicate token is generated and assigned
        retry = True
        while(retry):
            retry = False
            try:
                self.validate_unique()
            except ValidationError as e:
                if 'token' in e.message_dict:
                    self.token = secrets.token_urlsafe(32)
                    retry = True
        return self

    def get_signed_token(self):
        """Return the token as a url-safe signed string."""
        return Signer(sep='/', salt=SALT).sign(self.token)

    def is_expired(self):
        """Returns whether given token is expired."""
        return timezone.now().date() > expiration_date

    objects = NewUserManager()

    class Meta:
        verbose_name =_('new user')
        verbose_name_plural = _('new users')
        default_permissions = ()
