from django import forms
from vkk.workhours.models import (
    GeneralCosts, DepartmentDate, DepartmentCosts,
    SalaryLevelDate, SalaryLevelCosts, ReceiptTemplate,
    Department, SalaryLevel
)
from vkk.workhours.forms import CustomDateInput

class DepartmentCostForm(forms.ModelForm):
    class Meta:
        model = DepartmentDate
        fields = ['date']
        widgets = {
            'date': CustomDateInput(),
        }
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Collect all associated departments to create fields
        self._departments = Department.objects.all().order_by('name')
        field_class = DepartmentCosts.equivalents_per_hour.field.formfield
        for department in self._departments:
            self.fields[
                'dep_' + str(department.accounting_entry)
            ] = field_class(
                label=department.name
            )

        # Look for model instance and fill associated fields accordingly
        if self.instance is not None:
            department_cost_instances = DepartmentCosts.objects.filter(
                start=self.instance
            ).select_related('department')
            for department_cost_instance in department_cost_instances:
                self.fields[
                    'dep_' + str(
                        department_cost_instance.department.accounting_entry
                    )
                ].initial = department_cost_instance.equivalents_per_hour
    
    def save(self, commit=True):
        # save instance
        super().save(commit)
        # save all associated instances
        if self.is_valid() and self.has_changed():
            department_costs_list = []
            for department in self._departments:
                    equivalents_per_hour = self.cleaned_data.get(
                        'dep_' + str(department.accounting_entry)
                    )
                    department_costs_list.append(
                        DepartmentCosts(
                            department=department,
                            start=self.instance,
                            equivalents_per_hour=equivalents_per_hour,
                        )
                    )
            DepartmentCosts.objects.bulk_create(
                department_costs_list,
                update_conflicts=True,
                update_fields=['equivalents_per_hour'],
                unique_fields=['department_id', 'start_id']
            )
        return self.instance

class SalaryLevelCostForm(forms.ModelForm):
    class Meta:
        model = SalaryLevelDate
        fields = ['date']
        widgets = {
            'date': CustomDateInput(),
        }
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Collect all associated departments to create fields
        self._salary_levels = SalaryLevel.objects.all()
        field_class = SalaryLevelCosts.brutto_per_hour.field.formfield
        for salary_level in self._salary_levels:
            self.fields[
                'sal_' + str(salary_level.salary_code)
            ] = field_class(
                label=salary_level.salary_code
            )

        # Look for model instance and fill associated fields accordingly
        if self.instance is not None:
            salary_level_cost_instances = SalaryLevelCosts.objects.filter(
                start=self.instance
            ).select_related('salary_level')
            for salary_level_cost_instance in salary_level_cost_instances:
                self.fields[
                    'sal_' + str(
                        salary_level_cost_instance.salary_level.salary_code
                    )
                ].initial = salary_level_cost_instance.brutto_per_hour
    
    def save(self, commit=True):
        # save instance
        super().save(commit)
        # save all associated instances
        if self.is_valid() and self.has_changed():
            salary_level_costs_list = []
            for salary_level in self._salary_levels:
                    brutto_per_hour = self.cleaned_data.get(
                        'sal_' + str(salary_level.salary_code)
                    )
                    salary_level_costs_list.append(
                        SalaryLevelCosts(
                            salary_level=salary_level,
                            start=self.instance,
                            brutto_per_hour=brutto_per_hour,
                        )
                    )
            SalaryLevelCosts.objects.bulk_create(
                salary_level_costs_list,
                update_conflicts=True,
                update_fields=['brutto_per_hour'],
                unique_fields=['salary_level_id', 'start_id']
            )
        return self.instance