from django import forms
from decimal import Decimal
from django.db.models import Sum, F
from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.translation import gettext_lazy as _
from vkk.workhours import models
from vkk.generic.forms import CustomDateInput


class ReceiptForm(forms.ModelForm):
    class Meta:
        model = models.Receipt
        fields = ['start', 'end', 'receipt_number', 'buper']
        widgets = {
            'start': CustomDateInput,
            'end': CustomDateInput,
        }

    class Media:
        js = ('scripts/receipts.js',)

    def __init__(self, *args, project=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.project = project
        self.department = self.project.department
        self.general_costs = None
        self.department_costs = None
        self.project_funded_staff_date = None
        self.project_funded_staff = None
        self.salary_level_date = None
        self.salary_costs = None
        self.salary_costs_annotated1 = None
        self.salary_costs_annotated2 = None
        self.data_dict = None

    def set_and_clean_general_costs(self, start, end):
        if models.GeneralCosts.objects.filter(start__gt=start, start__lte=end,).exists():
            raise ValidationError(
                _('General cost records are ambiguous.'),
                code='ambiguous_general_costs'
            )
        try:
            self.general_costs = models.GeneralCosts.objects.filter(
                start__lte=start
            ).latest('start')
        except models.GeneralCosts.DoesNotExist:
            raise ValidationError(
                _('No valid general cost record found'),
                code='no_general_costs'
            )

    def set_and_clean_department_costs(self, start, end):
        if models.DepartmentCosts.objects.filter(
            department=self.department,
            start__date__gt=start,
            start__date__lte=end,
        ).exists():
            raise ValidationError(
                _('Department cost records are ambiguous.'),
                code='ambiguous_department_costs'
            )
        try:
            self.department_costs = models.DepartmentCosts.objects.filter(
                department=self.department,
                start__date__lte=start,
            ).latest('start__date')
        except models.DepartmentCosts.DoesNotExist:
            raise ValidationError(
                _('No valid department cost record found'),
                code='no_department_costs'
            )

    def set_and_clean_project_funded_staff(self, start, end):
        if models.ProjectFundedStaffDate.objects.filter(
            project=self.project,
            date__gt=start,
            date__lte=end,
        ).exists():
            raise ValidationError(
                _('Project funded staff records are ambiguous.'),
                code='ambiguous_staff_costs'
            )
        try:
            self.project_funded_staff_date = models.ProjectFundedStaffDate.objects.filter(
                project=self.project,
                date__lte=start,
            ).latest('date')
        except models.ProjectFundedStaffDate.DoesNotExist:
            # This is valid behaviour, no exception needed
            pass
        if self.project_funded_staff_date is not None:
            self.project_funded_staff = models.ProjectFundedStaff.objects.filter(
                start=self.project_funded_staff_date
            )

    def set_and_clean_salary_level(self,  start, end):
        if models.SalaryLevelDate.objects.filter(
            date__gt=start,
            date__lte=end,
        ).exists():
            raise ValidationError(
                _('Salary Level records are ambiguous.'),
                code='ambiguous_salary_costs'
            )
        try:
            self.salary_level_date = models.SalaryLevelDate.objects.filter(
                date__lte=start
            ).latest('date')
        except:
            raise ValidationError(
                _('No valid salary level cost records found'),
                code='no_salary_costs'
            )
        if self.salary_level_date is not None:
            self.salary_costs = models.SalaryLevelCosts.objects.filter(
                start=self.salary_level_date
            )

    def check_peroid_overlap(self, start, end):
        # check for matching period
        if not models.Period.objects.filter(start=start, end=end).exists():
            raise ValidationError(
                _('End and Start do not match up with given periods'),
                code='period_ambiguous'
            )

    def check_closed_periods(self, start, end):
        assignments_not_closed = models.ProjectAssignment.objects.filter(
            project=self.project
        ).exclude(
            periodclosure__period__start=start,
            periodclosure__period__end=end,
            periodclosure__is_closed_manager=True,
            periodclosure__is_closed_contributor=True,
        )
        if assignments_not_closed.exists():
            raise ValidationError(
                _('Some contributors or project managers did not close their work hour inputs'),
                code='periods_not_closed'
            )

    def set_and_clean_workhours(self, start, end):
        if self.salary_costs is not None:
            # Specify constraints for aggregations
            agg1 = self.salary_costs.filter(
                salary_level__projectassignment__project=self.project,
                salary_level__projectassignment__workhours__day__gte=start,
                salary_level__projectassignment__workhours__day__lte=end,
            )
            # Annotate work hours
            agg1 = agg1.annotate(
                workhours=Sum(
                    'salary_level__projectassignment__workhours__hours')
            )
            agg1 = agg1.annotate(
                costs=F('brutto_per_hour') *
                Sum('salary_level__projectassignment__workhours__hours')
            )

            self.salary_costs_annotated1 = agg1

            agg2 = self.salary_costs.filter(
                salary_level__projectassignment__project=self.project,
                salary_level__projectassignment__workhourscorrection__period__start=start,
                salary_level__projectassignment__workhourscorrection__period__end=end,
            )
            agg2 = agg2.annotate(
                workhours_correction=Sum(
                    'salary_level__projectassignment__workhourscorrection__ammount'
                )
            )
            agg2 = agg2.annotate(
                costs=F('brutto_per_hour') *
                Sum('salary_level__projectassignment__workhourscorrection__ammount')
            )

            self.salary_costs_annotated2 = agg2

    def clean(self):
        start = self.cleaned_data.get('start')
        end = self.cleaned_data.get('end')
        if start is not None and end is not None:
            self.set_and_clean_general_costs(start, end)
            self.set_and_clean_department_costs(start, end)
            self.set_and_clean_project_funded_staff(start, end)
            self.set_and_clean_salary_level(start, end)
            self.check_peroid_overlap(start, end)
            self.check_closed_periods(start, end)
            self.set_and_clean_workhours(start, end)
        return self.cleaned_data

    def save(self, commit=True):
        self.instance.project = self.project
        self.instance.data = self.to_json()
        return super().save(commit)

    def _project_dict(self):
        return {
            'project': {
                'invoice_number': self.project.invoice_number,
                'name': self.project.name,
                'contractor': self.project.contractor,
                'start': self.project.start,
                'end': self.project.end,
            }
        }

    def _department_dict(self):
        return {
            'department': {
                'name': self.department.name,
                'accounting_entry': self.department.accounting_entry,
                'invoice_number': self.department.invoice_number,
            }
        }

    def _general_costs_dict(self):
        return {
            'general_costs': {
                'start': self.general_costs.start,
                'costs': self.general_costs.costs,
            }
        }

    def _department_costs_dict(self):
        return {
            'department_costs': {
                'start': self.department_costs.start.date,
                'equivalents_per_hour': self.department_costs.equivalents_per_hour,
            }
        }

    def _project_funded_staff_dict(self):
        return {
            'project_funded_staff': {
                'start': self.project_funded_staff_date.date if
                self.project_funded_staff_date is not None else None,
                'hours_by_salary_level': {
                    entry.salary_level.salary_code: {
                        'hours': entry.hours,
                        'brutto_per_hour': self.salary_costs.get(
                            salary_level=entry.salary_level
                        ).brutto_per_hour,
                    } for entry in self.project_funded_staff or []
                },
                'hours_sum': sum([
                    entry.hours for entry in self.project_funded_staff or []
                ]),
            }
        }

    def _salary_costs_annotated_dict(self):
        summed_costs = []

        for entry in self.salary_costs:
            workhours, costs, code, brutto = None, None, None, None
            annotated1 = self.salary_costs_annotated1.filter(id=entry.id)
            if annotated1.exists():
                code = annotated1[0].salary_level.salary_code
                brutto = annotated1[0].brutto_per_hour
                workhours = annotated1[0].workhours
                costs = annotated1[0].costs

            annotated2 = self.salary_costs_annotated2.filter(id=entry.id)
            if annotated2.exists():
                code = annotated2[0].salary_level.salary_code
                brutto = annotated2[0].brutto_per_hour
                workhours += annotated2[0].workhours_correction
                costs += annotated2[0].costs

            if workhours is not None and costs is not None \
                    and code is not None and brutto is not None:
                summed_costs.append((workhours, costs, code, brutto))

        return {
            'salary_costs_annotated': {
                'start': self.salary_level_date.date,
                'salary_levels': {
                    entry[2]: {
                        'brutto_per_hour': entry[3],
                        'hours': entry[0],
                        'costs': entry[1],
                    } for entry in summed_costs
                },
                'hours_sum': sum([
                    entry[0] for entry in summed_costs
                ]),
                'costs_sum': sum([
                    entry[1] for entry in summed_costs
                ]),
            }
        }

    def to_data_dict(self):
        self.data_dict = self._project_dict() \
            | self._department_dict() \
            | self._general_costs_dict() \
            | self._department_costs_dict() \
            | self._project_funded_staff_dict() \
            | self._salary_costs_annotated_dict()

        self.data_dict['department_costs'].update(
            {
                'salary_costs': self.department_costs.equivalents_per_hour
                * (self.data_dict['project_funded_staff']['hours_sum']
                   + self.data_dict['salary_costs_annotated']['hours_sum']),
            }
        )
        self.data_dict['general_costs'].update(
            {
                'total': (self.data_dict['project_funded_staff']['hours_sum']
                          + self.data_dict['salary_costs_annotated']['hours_sum'])
                * self.data_dict['general_costs']['costs'],
            }
        )

        return self.data_dict

    def to_json(self):
        return CostumJSONEncoder().encode(self.to_data_dict())


class CostumJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return str(round(o, 2)).replace(".", ",")
        else:
            return super().default(o)


class ReceiptTemplateSelectForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        queryset = models.ReceiptTemplate.objects.all().order_by('-start')
        self.fields["receipt_template"] = forms.ModelChoiceField(
            queryset,
            empty_label=None,
            label=_('Receipt Template')
        )
