from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView, RedirectView
from django.urls import reverse
from django.db.models import Max
from django.utils import timezone
from vkk.generic.views import CustomListView, CustomSuccessView
from vkk.users.models import User
from vkk.workhours.models import Period
from .mixins import ProjectManagerRequiredMixin
from .forms import AssigneeForm, YearSelectForm, AssgineeClosureForm


class AllProjectYearRedirectView(ProjectManagerRequiredMixin, RedirectView):
    def get_redirect_url(self, *args, **kwargs):
        # Catches Period Selection from GET
        if 'year' in self.request.GET:
            form = YearSelectForm(data=self.request.GET)
            if form.is_valid():
                year = int(form.cleaned_data['year'])
        else:
            year = Period.objects.aggregate(Max('start'))['start__max'].year
        return reverse('vkk:workhours:allprojects:overview', kwargs={'year': year})


class AllProjectAssigneesOverView(ProjectManagerRequiredMixin, CustomListView):
    model = User
    fields = ['first_name', 'last_name', 'email']
    action_options = {'update': _('Details')}
    ordering = ['last_name', 'first_name']
    template_name = 'vkk/workhours/allprojects/overview.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form"] = YearSelectForm(
            initial={'year': self.kwargs['year']}
        )
        return context

    def get_queryset(self):
        query_set = super().get_queryset().filter(
            projectassignment__project__projectmanager__manager=self.request.user,
            projectassignment__project__projectmanager__end__gte=timezone.now().date(),
        ).distinct()
        return query_set


class AssigneeUpdate(ProjectManagerRequiredMixin, FormView):
    form_class = AssigneeForm
    template_name = 'vkk/workhours/allprojects/assignee_update.html'

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update({
            'assignee': self.kwargs['pk'],
            'manager': self.request.user,
            'year': self.kwargs['year'],
        })
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["assignee"] = User.objects.get(id=self.kwargs['pk'])
        return context

    def get_success_url(self):
        return reverse(
            'vkk:workhours:allprojects:update_success', kwargs={'year': self.kwargs['year']},
        )

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)


class AssigneeClosureView(ProjectManagerRequiredMixin, FormView):
    form_class = AssgineeClosureForm
    template_name = 'vkk/workhours/contributor/closure.html'

    def post(self, request, *args, **kwargs):
        form = self.form_class(
            data={
                'period': self.kwargs['period_pk'],
                'assignee': self.kwargs['assignee_pk'],
                'manager': self.request.user
            },
        )
        if form.is_valid:
            form.save()
            return self.form_valid(form)
        else:
            return self.handle_no_permission()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.pop('form')
        return context

    def get_success_url(self):
        url = reverse(
            'vkk:workhours:allprojects:closure_success',
            args=[self.kwargs['year'], self.kwargs['assignee_pk']],
        )
        return url


class AssigneeUpdateSuccessView(ProjectManagerRequiredMixin, CustomSuccessView):
    pass


class AssigneeClosureSuccessView(ProjectManagerRequiredMixin, CustomSuccessView):

    def get_success_url(self):
        url = reverse(
            'vkk:workhours:allprojects:update',
            args=[self.kwargs['year'], self.kwargs['assignee_pk']],
        )
        return url
