from django.views.generic import RedirectView, FormView, ListView, TemplateView
from django.utils import timezone
from django.db.models import Sum, Q
from django.http import Http404
from django.urls import reverse
from django.shortcuts import get_object_or_404, redirect
from django.forms import Form, modelform_factory
from vkk.generic.views import CustomSuccessView
from vkk.generic.mixins import OnSuccessMixin
from vkk.workhours.models import Period, WorkHours, PeriodClosure, ProjectAssignment, WorkHoursCorrection
from vkk.workhours.forms import PeriodSelectForm
from .forms import WorkhourCalendarForm
from .mixins import ContributorRequiredMixin, ContributorPeriodSelectorMixin


class PeriodRedirect(ContributorRequiredMixin, RedirectView):
    """ Redirects a contributor to a period."""

    def get_redirect_url(self, *args, **kwargs):
        # Catches Period Selection from GET
        if 'period' in self.request.GET:
            now = timezone.now()
            query_set = Period.objects.filter(
                Q(dead_line__gte=now) |
                Q(workhours__project_assignment__pk=self.kwargs['assignment_pk']) |
                Q(periodclosure__project_assignment__pk=self.kwargs['assignment_pk'])
            ).distinct()
            form = PeriodSelectForm(query_set, data=self.request.GET)
            if form.is_valid():
                period = form.cleaned_data.get("period")
            else:
                raise Http404()
        # Catches Period no given
        elif 'period_pk' not in kwargs:
            period = Period.objects.latest(create=True)
        # Looks up Period
        else:
            period = get_object_or_404(Period, pk=kwargs['period_pk'])

        return reverse(
            'vkk:workhours:contributor:calendar',
            args=[kwargs['assignment_pk'], period.pk]
        )


class WorkhoursView(ContributorRequiredMixin, ContributorPeriodSelectorMixin, FormView):
    """Displays a calendar for entering work hours."""
    form_class = WorkhourCalendarForm
    template_name = 'vkk/workhours/contributor/workhours.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["project"] = ProjectAssignment.objects.select_related('project').get(
            pk=self.kwargs['assignment_pk']
        ).project
        return context

    def dispatch(self, request, *args, **kwargs):
        period = Period.objects.filter(
            pk=self.kwargs['period_pk'],
            dead_line__lt=timezone.now()
        )
        if period.exists():
            return redirect('vkk:workhours:contributor:summary', **self.kwargs)

        closure = PeriodClosure.objects.filter(
            project_assignment__pk=self.kwargs['assignment_pk'],
            period__pk=self.kwargs['period_pk']
        )
        if closure.exists() and (closure[0].is_closed_contributor or closure[0].is_closed_manager):
            return redirect('vkk:workhours:contributor:summary', **self.kwargs)
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update({
            'period_pk': self.kwargs['period_pk'],
            'assignment_pk': self.kwargs['assignment_pk'],
        })
        return kwargs

    def get_success_url(self):
        return reverse(
            'vkk:workhours:contributor:summary',
            args=[
                self.kwargs['assignment_pk'],
                self.kwargs['period_pk']
            ]
        )

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)


class SummaryView(ContributorRequiredMixin, ContributorPeriodSelectorMixin, ListView):
    model = WorkHours
    template_name = 'vkk/workhours/contributor/summary.html'

    def get_queryset(self):
        query_set = super().get_queryset()
        query_set = query_set.filter(
            project_assignment__pk=self.kwargs['assignment_pk'],
            period__pk=self.kwargs['period_pk']
        )
        return query_set.order_by('day')

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        query_set = context["object_list"]

        correction = WorkHoursCorrection.objects.filter(
            project_assignment__pk=self.kwargs['assignment_pk'],
            period__pk=self.kwargs['period_pk']
        )
        context["correction"] = correction

        context.update(query_set.aggregate(Sum('hours')))
        if context["hours__sum"] is None:
            context["hours__sum"] = 0
        if correction.exists():
            context["hours__sum"] += correction[0].ammount

        context["fields"] = ['day', 'hours']

        closure = PeriodClosure.objects.filter(
            project_assignment__pk=self.kwargs['assignment_pk'],
            period__pk=self.kwargs['period_pk']
        )
        if closure.exists() and (closure[0].is_closed_contributor or closure[0].is_closed_manager):
            context['closure'] = True
        else:
            context['closure'] = False

        context['project'] = ProjectAssignment.objects.select_related('project').get(
            pk=self.kwargs['assignment_pk']
        ).project

        return context


class ClosureView(ContributorRequiredMixin, OnSuccessMixin, FormView):
    template_name = 'vkk/workhours/contributor/closure.html'
    form_class = Form
    on_success = 'closure_success'

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        model_form = modelform_factory(
            model=PeriodClosure,
            fields='__all__',
        )(
            data={
                'period': self.kwargs['period_pk'],
                'project_assignment': self.kwargs['assignment_pk'],
                'is_closed_contributor': True,
            },
        )
        if form.is_valid and model_form.is_valid:
            model_form.save()
            return self.form_valid(form)
        else:
            return self.handle_no_permission()


class ClosureSuccessView(ContributorRequiredMixin, CustomSuccessView):
    template_name = 'vkk/workhours/contributor/closure_success.html'
    model = PeriodClosure
    on_success = 'summary'
