from datetime import datetime, timedelta, date
from django.db import models
from django.conf import settings
from django.utils.timezone import template_localtime, now, make_aware
from django.utils.formats import date_format
from django.utils.text import format_lazy
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator, MaxValueValidator
from django.utils.translation import gettext_lazy as _

# Create your models here.


class Department(models.Model):
    """Model for Departments"""
    name = models.CharField(
        verbose_name=_('name'),
        max_length=255,
    )
    accounting_entry = models.CharField(
        verbose_name=_('accounting entry'),
        max_length=7,
        unique=True
    )
    invoice_number = models.IntegerField(
        null=True,
        verbose_name=_('invoice number'),
    )

    def __str__(self):
        return str(self.name)

    class Meta:
        verbose_name = _('department')
        verbose_name_plural = _('departments')
        default_permissions = ()


class Project(models.Model):
    """Model to track projects"""
    invoice_number = models.IntegerField(
        verbose_name=_('invoice number'),
        unique=True,
        validators=[
            MinValueValidator(20000000),
            MaxValueValidator(39999999),
        ],
        help_text=_(
            'Identifier for invoice as number between 20000000 and 39999999.')
    )
    name = models.CharField(
        verbose_name=_('name'),
        max_length=255
    )
    contractor = models.CharField(
        verbose_name=_('contractor'),
        max_length=255
    )
    start = models.DateField(
        verbose_name=_('start date'),
    )
    end = models.DateField(
        verbose_name=_('end date'),
    )
    department = models.ForeignKey(
        Department,
        on_delete=models.SET_NULL,
        null=True,
        verbose_name=_('department')
    )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        if (not exclude or 'end' not in exclude) and self.start > self.end:
            raise ValidationError({
                'end': ValidationError(
                    _('The end date must be after the start date.'),
                    code='invalid_dates'
                )
            })

    def __str__(self):
        return str(self.name)

    class Meta:
        verbose_name = _('project')
        verbose_name_plural = _('projects')
        default_permissions = ()
        constraints = [
            models.CheckConstraint(
                check=models.Q(start__lte=models.F('end')),
                name='runtime_check_projekt'
            ),
            models.CheckConstraint(
                check=models.Q(invoice_number__gte=20000000) & models.Q(
                    invoice_number__lte=39999999),
                name='invoice_number_intervall_check'
            )
        ]


class ProjectManager(models.Model):
    """A model for tracking user in role of project managers."""
    manager = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        verbose_name=_('manager'),
        on_delete=models.CASCADE,
    )
    project = models.ForeignKey(
        to=Project,
        on_delete=models.CASCADE,
        verbose_name=_('project')
    )
    start = models.DateField(
        verbose_name=_('start date'),
    )
    end = models.DateField(
        verbose_name=_('end date')
    )
    is_proxy = models.BooleanField(
        verbose_name=_('proxy'),
        default=False,
        help_text=_(
            'Designates whether management rights are given to the user'
            'for administrative uses only.'
        ),
    )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        if (not exclude or 'end' not in exclude) and self.start > self.end:
            raise ValidationError({
                'end': ValidationError(
                    _('The end date must be after the start date.'),
                    code='invalid_dates'
                )
            })

    class Meta:
        verbose_name = _('project manager')
        verbose_name_plural = _('project managers')
        default_permissions = ()
        constraints = [
            models.CheckConstraint(
                check=models.Q(start__lte=models.F('end')),
                name='runtime_check_manager'
            ),
            models.UniqueConstraint(
                fields=['manager', 'project'],
                name='unique_constraint_project_manager',
            )
        ]


class SalaryLevel(models.Model):
    """Model to track salary levels."""
    salary_code = models.CharField(
        verbose_name=_('salary code'),
        max_length=15,
        unique=True
    )

    class Meta:
        verbose_name = _('salary level')
        verbose_name_plural = _('salary levels')
        default_permissions = ()

    def __str__(self):
        return str(self.salary_code)


class ProjectAssignment(models.Model):
    """A model to track users assigned to projects."""
    contributor = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name=_('contributor'),
    )
    salary_level = models.ForeignKey(
        to=SalaryLevel,
        on_delete=models.RESTRICT,
        verbose_name=_('salary level'),
    )
    project = models.ForeignKey(
        to=Project,
        on_delete=models.CASCADE,
        verbose_name=_('project')
    )

    class Meta:
        verbose_name = _('project assignment')
        verbose_name_plural = _('project assignments')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['contributor', 'salary_level', 'project'],
                name='unique_constraint_project_assignment',
            )
        ]


class PeriodManager(models.Manager):
    """Manager for Periods"""

    def latest(self, *fields, create=False):
        """Extend latest to allow automatic creation of a valid periods."""
        if create:
            latest = None
            today = now().date()
            try:
                latest = super().latest(*fields)
            except Period.DoesNotExist:
                start = today.replace(day=1)
            if latest and latest.end < today:
                start = latest.end + timedelta(days=1)
            if not latest or latest.end < today:
                end = (
                    (start.replace(day=1) + timedelta(days=32)).replace(day=1)
                ) - timedelta(days=1)
                dead_line = make_aware(
                    datetime.combine(
                        (end + timedelta(days=32)).replace(day=1),
                        datetime.min.time()
                    ) - timedelta(microseconds=1)
                )
                latest = self.create(
                    start=start,
                    end=end,
                    dead_line=dead_line,
                    dead_line_final=dead_line
                )
            if latest.end < today:
                return self.latest(*fields, create=create)
            else:
                return latest
        else:
            return super().latest(*fields)


class Period(models.Model):
    """Model to track time periods."""
    start = models.DateField(
        verbose_name=_('start date')
    )
    end = models.DateField(
        verbose_name=_('end date')
    )
    dead_line = models.DateTimeField(
        verbose_name=_('dead line')
    )
    dead_line_final = models.DateTimeField(
        verbose_name=_('final dead line')
    )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        if (not exclude or 'end' not in exclude) and self.start > self.end:
            raise ValidationError({
                'end': ValidationError(
                    _('The end date must be after the start date.'),
                    code='invalid_dates'
                )
            })
        if (not exclude or 'dead_line' not in exclude) and self.end > self.dead_line.date():
            raise ValidationError({
                'dead_line': ValidationError(
                    _('The dead line must be after the end date.'),
                    code='invalid_dead_line'
                )
            })
        if (not exclude or 'dead_line_final' not in exclude) and self.dead_line.date() > self.dead_line_final.date():
            raise ValidationError({
                'dead_line_final': ValidationError(
                    _('The final dead line cannot be before the dead line.'),
                    code='invalid_dead_line_final'
                )
            })

    def __str__(self):
        start = date_format(
            self.start, format='SHORT_DATE_FORMAT', use_l10n=True)
        end = date_format(self.end, format='SHORT_DATE_FORMAT', use_l10n=True)
        dead_line = date_format(
            template_localtime(self.dead_line),
            format='SHORT_DATETIME_FORMAT', use_l10n=True
        )
        return '%s - %s ( %s )' % (start, end, dead_line)

    objects = PeriodManager()

    class Meta:
        verbose_name = _('period')
        verbose_name_plural = _('periods')
        default_permissions = ()
        constraints = [
            models.CheckConstraint(
                check=models.Q(start__lte=models.F('end')),
                name='runtime_check_period'
            ),
            models.CheckConstraint(
                check=models.Q(end__lte=models.F('dead_line')),
                name='dead_line_check_period'
            ),
            models.CheckConstraint(
                check=models.Q(dead_line__lte=models.F('dead_line_final')),
                name='dead_line_final_check_period'
            )
        ]
        get_latest_by = 'start'


class PeriodClosure(models.Model):
    """
    A model to track whether a user, assigned to a project, has permission to
    add, change or delete work hour records for a given period. 
    """
    period = models.ForeignKey(
        to=Period,
        on_delete=models.CASCADE,
        verbose_name=('period')
    )
    project_assignment = models.ForeignKey(
        to=ProjectAssignment,
        on_delete=models.CASCADE,
        verbose_name=('project assignment')
    )
    is_closed_contributor = models.BooleanField(
        default=False,
        verbose_name=('closed for contributor')
    )
    is_closed_manager = models.BooleanField(
        default=False,
        verbose_name=('closed for manager')
    )

    class Meta:
        verbose_name = _('period')
        verbose_name_plural = _('periods')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['period', 'project_assignment'],
                name='unique_constraint_period_closure',
            )
        ]


class WorkHours(models.Model):
    """Model to keep track of work hours."""
    project_assignment = models.ForeignKey(
        to=ProjectAssignment,
        on_delete=models.RESTRICT,
        verbose_name=_('project assignment'),
    )
    period = models.ForeignKey(
        to=Period,
        on_delete=models.RESTRICT,
        verbose_name=_('period')
    )
    day = models.DateField(
        verbose_name=_('work day'),
        help_text=_('Date of work day.')
    )
    hours = models.DecimalField(
        max_digits=4,
        decimal_places=2,
        validators=[
            MinValueValidator(0.01),
            MaxValueValidator(24),
        ],
        verbose_name=_('work hours'),
        help_text=_('Number of hours worked with '
                    'precision up to two decimal places.'),
    )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        if not exclude or 'day' not in exclude:
            if self.period.start < self.day or self.day > self.period.end:
                raise ValidationError({
                    'day': ValidationError(
                        _('Work day is not between the start and end'
                          'date of the given period.'),
                        code='invalid_day'
                    )
                })

    class Meta:
        verbose_name = _('work hours')
        verbose_name_plural = _('work hours')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['project_assignment', 'day'],
                name='unique_constraint_work_hours',
            ),
            models.CheckConstraint(
                check=models.Q(hours__gt=0.0),
                name='min_work_hours'
            ),
            models.CheckConstraint(
                check=models.Q(hours__lte=24.0),
                name='max_work_hours'
            ),
        ]


class WorkHoursCorrection(models.Model):
    project_assignment = models.ForeignKey(
        to=ProjectAssignment,
        on_delete=models.RESTRICT,
        verbose_name=_('project assignment'),
    )
    period = models.ForeignKey(
        to=Period,
        on_delete=models.RESTRICT,
        verbose_name=_('period')
    )
    ammount = models.DecimalField(
        max_digits=5,
        decimal_places=2,
        help_text=_('Ammount by which the total number of workhours'
                    'for a given project assignment and period should be corrected.'),
    )

    class Meta:
        verbose_name = _('work hours correction')
        verbose_name_plural = _('work hours correction')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['project_assignment', 'period'],
                name='unique_constraint_work_hours_correction',
            ),
        ]


###### Export ######
class ProjectFundedStaffDate(models.Model):
    date = models.DateField(
        verbose_name=_('start date'),
    )
    project = models.ForeignKey(
        'Project',
        verbose_name=_('project'),
        on_delete=models.CASCADE,
    )

    def __str__(self):
        date = date_format(
            self.date, format='SHORT_DATE_FORMAT', use_l10n=True)
        verbose = self._meta.verbose_name
        return f"{verbose}: {date}"

    class Meta:
        verbose_name = _('project funded staff interval')
        verbose_name_plural = _('project funded staff intervals')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['date', 'project'],
                name='unique_constraint_project_funded_staff_date',
            ),
        ]


class ProjectFundedStaff(models.Model):
    salary_level = models.ForeignKey(
        'SalaryLevel',
        verbose_name=_('salary level'),
        on_delete=models.CASCADE,
    )
    start = models.ForeignKey(
        'ProjectFundedStaffDate',
        verbose_name=_('start date'),
        on_delete=models.CASCADE
    )
    hours = models.DecimalField(
        max_digits=8,
        decimal_places=2,
    )

    class Meta:
        verbose_name = _('project funded staff')
        verbose_name_plural = _('project funded staff')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['salary_level', 'start'],
                name='unique_constraint_project_funded_staff',
            ),
        ]


class GeneralCosts(models.Model):
    start = models.DateField(
        verbose_name=_('start date'),
    )
    costs = models.DecimalField(
        verbose_name=_('general costs'),
        max_digits=8,
        decimal_places=2,
    )

    def __str__(self):
        start = date_format(
            self.start, format='SHORT_DATE_FORMAT', use_l10n=True)
        verbose = self._meta.verbose_name
        return f"{verbose}: {start}"

    class Meta:
        verbose_name = _('general costs')
        verbose_name_plural = _('general costs')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['start'],
                name='unique_constraint_general_costs',
            ),
        ]


class DepartmentDate(models.Model):
    date = models.DateField(
        verbose_name=_('start date'),
    )

    def __str__(self):
        date = date_format(
            self.date, format='SHORT_DATE_FORMAT', use_l10n=True)
        verbose = self._meta.verbose_name
        return f"{verbose}: {date}"

    class Meta:
        verbose_name = _('department interval')
        verbose_name_plural = _('department intervals')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['date'],
                name='unique_constraint_department_date',
            ),
        ]


class DepartmentCosts(models.Model):
    department = models.ForeignKey(
        'Department',
        verbose_name=_('department'),
        on_delete=models.CASCADE
    )
    start = models.ForeignKey(
        'DepartmentDate',
        verbose_name=_('start date'),
        on_delete=models.CASCADE
    )
    equivalents_per_hour = models.DecimalField(
        verbose_name=_('full time equivalents per hour'),
        max_digits=8,
        decimal_places=2,
    )

    def __str__(self):
        return f"{self._meta.verbose_name}: {self.pk}"

    class Meta:
        verbose_name = _('department costs')
        verbose_name_plural = _('department costs')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['department', 'start'],
                name='unique_constraint_department_costs',
            ),
        ]


class SalaryLevelDate(models.Model):
    date = models.DateField(
        verbose_name=_('start date'),
    )

    def __str__(self):
        date = date_format(
            self.date, format='SHORT_DATE_FORMAT', use_l10n=True)
        verbose = self._meta.verbose_name
        return f"{verbose}: {date}"

    class Meta:
        verbose_name = _('salary interval')
        verbose_name_plural = _('salary intervals')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['date'],
                name='unique_constraint_salary_level_date',
            ),
        ]


class SalaryLevelCosts(models.Model):
    salary_level = models.ForeignKey(
        'SalaryLevel',
        verbose_name=_('salary level'),
        on_delete=models.CASCADE,
    )
    start = models.ForeignKey(
        'SalaryLevelDate',
        verbose_name=_('start date'),
        on_delete=models.CASCADE
    )
    brutto_per_hour = models.DecimalField(
        verbose_name=_('brutto staff costs per hour'),
        max_digits=8,
        decimal_places=2,
    )

    def __str__(self):
        return f"{self._meta.verbose_name}: {self.pk}"

    class Meta:
        verbose_name = _('salary level costs')
        verbose_name_plural = _('salary level costs')
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=['salary_level', 'start'],
                name='unique_constraint_salary_level_costs',
            ),
        ]


class ReceiptTemplate(models.Model):
    start = models.DateField(
        verbose_name=_('start date'),
    )
    data = models.JSONField()

    def __str__(self):
        return f"{self._meta.verbose_name}: {self.start}"

    class Meta:
        default_permissions = ()
        verbose_name = _('receipt template')
        verbose_name_plural = ('receipt templates')
        constraints = [
            models.UniqueConstraint(
                fields=['start'],
                name='unique_constraint_receipt_template_start',
            ),
        ]


class Receipt(models.Model):
    start = models.DateField(
        verbose_name=_('start date'),
    )
    end = models.DateField(
        verbose_name=_('end date'),
    )
    create_time = models.DateTimeField(
        verbose_name=_('time booked'),
        auto_now=False,
        auto_now_add=True,
    )
    project = models.ForeignKey(
        'Project',
        verbose_name=_('project'),
        on_delete=models.CASCADE,
    )
    receipt_number = models.IntegerField(
        verbose_name=_('receipt number'),
    )
    buper = models.IntegerField(
        verbose_name=_('buper'),
    )

    data = models.JSONField()

    class Meta:
        default_permissions = ()
        verbose_name = _('receipt')
        verbose_name_plural = ('receipts')
        constraints = [
            models.CheckConstraint(
                check=models.Q(start__lte=models.F('end')),
                name='interval_check_receipt'
            ),
        ]
